/* 
========================================
   ** SupCom MaxTool - Jan Van Der Weg **
          www.janvanderweg.com

	Many thanks to:
		Chris Cookson for his scripts, they taught me a lot!
		der_ton from doom3world, without his scripts correct bone creation would have taken a lot longer!
		GH33DA for explaining what the transform matrixes (matrices?) are!
	
	SupCom Maxtool:
	v0.6
	-Fixed some bone import issues
	-Added export functionality.
	v0.49a
	-Fixed bone creation. It should all be correct now. 
	-Optimized a lot of the script and fixed some potential issues.
	-Still some small things to add but the base import functions are finished.
	v0.4a
	-Planning on merging scm/sca import and export into a single maxscript: SupCom MaxTool
	-UI layout change:
		-Added option to load reference points as dummies or bones.
		-Added progress bar
	-Bones are now max bone objects.
	-Bone location is still off with the spiderbot. I think this is still a bug :(
	-Apply sca button has no functionality yet, sca import script has not been merged in yet.
========================================
*/
utility SCMTool "SupCom MaxTool" silentErrors:false
(
	--Rollout
	group "About:"
	(
		bitmap the_bmp fileName:"supcom.bmp" 
		label Label1 "SupCom MaxTool v0.49a"
		label Label2 "www.janvanderweg.com"
	)
	
	group "Import:"
	(
		--label Label3 "Load bones..."
		--radiobuttons radiobones labels:#("Only weighted", "All bones")
		--checkbox fixrotcheck "Fix model rotation" tristate:1
		checkbox loadrefpoints "Load ref. bones as:" tristate:1
		radiobuttons refpointsas labels:#("Bones","Dummies")
		checkbox fixrotcheck "Fix model rotation"

		button importButton "Import .SCM"	tooltip:"Import a Supreme Command An RTS From Chris Taylor model... ^_^"
		button applySCA "Apply .SCA" tooltip:"Load and apply animations from an .sca file... ^_^"

	)
	group "Export:"
	(
		button exportSCM "Export .SCM" tooltip:"Export a Supreme Commander An RTS From Chris Taylor model... ^_^"
	)
	
	---------------
	-- Utilities --
	---------------
	-- Replacement for doesFileExist which seems to give false positives
	fn DoesFileReallyExist filename=
	(
		local temp = fopen filename "rb"
		if temp != undefined then
		(
			fclose temp
			true
		)
		else
		(
			false
		)
	)
	-- Read a string of fixedLen chars
	fn ReadFixedString bstream fixedLen junk=
	(
		local str = ""
		for i = 1 to fixedLen do
		(
			local ch = ReadByte bstream #unsigned
			--format "Read string: %\n" ch
			if junk == true then
			(
				if ch == 0 then
				(
					str += bit.intAsChar ch
					-- Ignore junk
					SkipBytes bstream (fixedLen - i)
					exit
				)
				else 
				(
					str += bit.intAsChar ch
				)
			)
			--str += bit.intAsChar ch
			else if junk == false then
			(
				if ch == 0 then 
				(
					--replace with a tab "\t"
					str += "\t"

				)
				else
				(
					str += bit.intAsChar ch
				)
			)
		)
		str
	)
	--triangle struct
	struct SCM_Tris
	(
		triindex, --3 unsigned shorts 
		
		fn LoadFromStream bstream=
		(
			local x = Readshort bstream #unsigned
			local y = Readshort bstream #unsigned
			local z = Readshort bstream #unsigned
			triindex = [x, y, z]
		)
	)
	--vertex struct
	struct SCM_Vert
	(
		Position, --float vector(x,y,z) position relative to the local origion of the mesh
		--Tangent Space
		Normal, -- float vector(x,y,z)
		Tangent, -- float vector(x,y,z)
		Binormal, -- float vector(x,y,z)
		--supports two sets of uv coordinates (awesome!)
		uv1, --floatvector(u,v)
		uv2, --floatvector(u,v)
		boneindex, -- array of 4 chars - verts can be weighted to up to 4 bones. Supcom only expects boneindex[0] (non organic models)
	
		fn LoadFromStream bstream=
		(
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Position = [x, y, z]
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Normal = [x, y ,z]
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Tangent = [x, y, z]
				local x = ReadFloat bstream
				local y = ReadFloat bstream
				local z = ReadFloat bstream
				Binormal = [x, y, z]
				--hardcode the w component
				local u = ReadFloat bstream
				local v = ( 1 - ReadFloat bstream )
				uv1 = [u, v, 0]
				--hardcode the w component
				local u = ReadFloat bstream
				local v = ( 1 - ReadFloat bstream )
				uv2 = [u, v, 0]
				local x = Readbyte bstream #unsigned
				local y = Readbyte bstream #unsigned
				local z = Readbyte bstream #unsigned
				local w = Readbyte bstream #unsigned
				boneindex = [x, y, z, w]
		),
		fn DebugOut=
		(
			format "%\n" Position
		)	
	)
	--bone struct
	struct SCM_Bone
	(
	    RestPoseInverse, 	--4x4 Matrix with row major 
		Position,				--vector(x,z,y) pos relative to parent bone
		Rotation,				--quaternion(w,x,y,z) rot relative to parent bone
		nameoffset,				--long
		parentboneindex,		--long
		reserved0,				--long				
		reserved1,				--long
		Rotation2,
		rotquat,
			
		fn LoadFromStream bstream=
		(
			
			row1_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			row2_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			row3_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			row4_point3 = point3 ( ReadFloat bstream ) ( ReadFloat bstream ) ( ReadFloat bstream )
			w = ReadFloat bstream
			
			RestPoseInverse = matrix3 row1_point3 row2_point3 row3_point3 row4_point3
					
			x = ReadFloat bstream
			y = ReadFloat bstream
			z = ReadFloat bstream
			Position = [x, y, z]
			w = ReadFloat bstream
			x = ReadFloat bstream
			y = ReadFloat bstream
			z = ReadFloat bstream
			rotquat = Quat x y z w
			nameoffset = ReadLong bstream #unsigned
			parentboneindex = ( ReadLong bstream #unsigned + 1 )
			reserved0 = ReadLong bstream #unsigned
			reserved1 = ReadLong bstream #unsigned
		),
		fn DebugOut=
		(
			format "Restpose:\n"
			format "%\n" ( inverse Restposeinverse )
		)
			
	)
	struct SCM_Header
	(
		fourcc,			-- "MODL"
		version,		-- "5"
		boneoffset,		
		wbone,			--numer of bones influencing verts
		vertoffset,
		evertoffset,	--not used in supcom1.0, omitted
		vertcount,		--vertex count
		indexoffset,		
		indexcount,		--tricount = indexcount/3
		infooffset,
		infocount,
		totalbones,		--total amount of bones (weighted bones and tags)
		tricount,
	
		fn LoadFromStream bstream=
		(
			fourcc 		= ReadFixedString bstream 4 true
			version 	= ReadLong bstream #unsigned
			boneoffset 	= ReadLong bstream #unsigned
			wbone 		= ReadLong bstream #unsigned
			vertoffset 	= ReadLong bstream #unsigned
			evertoffset = ReadLong bstream #unsigned
			vertcount 	= ReadLong bstream #unsigned
			indexoffset = ReadLong bstream #unsigned
			indexcount 	= ReadLong bstream #unsigned
			infooffset 	= ReadLong bstream #unsigned
			infocount 	= ReadLong bstream #unsigned
			totalbones 	= ReadLong bstream #unsigned
			tricount 	= ( indexcount / 3 )
		),

		fn IsValid=
		(
			if fourcc != "MODL" then return false
			if version != 5 then
			(
				ShowError "Incorrect version! Found " + version + " but should be 5"
				return false
			)
			true
		),
		
		fn DebugOut=
		(
			format "SCM_Header\n----------\n"
			format "Marker: %, version: %\n" fourcc version
			format "Weighted bones: %, Total bones: %\n" wbone totalbones
			format "Vertex count: %, Trianlge count: %\n" vertcount ( indexcount / 3 )
			format "bOffset: %, vOffset: %, tOffset: %, iOffset %\n" boneoffset vertoffset indexoffset infooffset
			ok
		)
	)	
	--triangle struct
	struct e_SCM_Tris
	(
		triindex, --3 unsigned shorts 

		fn WriteTris bstream=
		(
			WriteShort bstream ( triindex[1] - 1 ) #unsigned 
			WriteShort bstream ( triindex[2] - 1 ) #unsigned
			WriteShort bstream ( triindex[3] - 1 ) #unsigned
		)		
	)
	--vertex struct
	struct e_SCM_Vert
	(
		Position, --float vector(x,y,z) position relative to the local origion of the mesh
		--Tangent Space
		Normal, -- float vector(x,y,z)
		Tangent, -- float vector(x,y,z)
		Binormal, -- float vector(x,y,z)
		--supports two sets of uv coordinates (awesome!)
		uv1, --floatvector(u,v)
		uv2, --floatvector(u,v)
		boneindex, -- array of 4 chars - verts can be weighted to up to 4 bones. Supcom only expects boneindex[0] (non organic models)
	
		fn WriteVert bstream=
		(
			--position
			WriteFloat bstream Position[1]
			WriteFloat bstream Position[2]
			WriteFloat bstream Position[3]
			--normal
			WriteFloat bstream Normal[1]
			WriteFloat bstream Normal[2]
			WriteFloat bstream Normal[3]
			--tangent
			WriteFloat bstream Tangent[1]
			WriteFloat bstream Tangent[2]
			WriteFloat bstream Tangent[3]
			--binormal
			WriteFloat bstream Binormal[1]
			WriteFloat bstream Binormal[2]
			WriteFloat bstream Binormal[3]
			--uv1
			WriteFloat bstream uv1[1]
			WriteFloat bstream ( 1 - uv1[2] )
			--uv2
			WriteFloat bstream uv2[1]
			WriteFloat bstream ( 1 - uv2[2] )
			--boneindex
			WriteByte bstream boneindex #unsigned
			WriteByte bstream 0 #unsigned
			WriteByte bstream 0 #unsigned
			WriteByte bstream 0 #unsigned
		),
		fn DebugOut=
		(
			format "Position: %\n" Position
			format "Normal: % Tanget: % Binormal: %\n" Normal Tanget Binormal
			format "UV1: % UV2: %2\n" uv1 uv2
			format "Boneindex: %\n" boneindex
		)	
	)
	--bone struct
	struct e_SCM_Bone
	(
	    RestPoseInverse1, 	--4x4 Matrix with row major
		RestPoseInverse2,	--row2 
		RestPoseInverse3,	--row3
		RestPoseInverse4,	--row4
		Position,				--vector(x,z,y) pos relative to parent bone
		Rotation,				--quaternion(w,x,y,z) rot relative to parent bone
		nameoffset,				--long
		parentboneindex,		--long
		reserved0,				--long				
		reserved1,				--long
		name,					--name
			
		fn WriteBone bstream=
		(
			--Rest Pose Inverse
			WriteFloat bstream RestPoseInverse1[1] 
			WriteFloat bstream RestPoseInverse2[1] 
			WriteFloat bstream RestPoseInverse3[1] 
			WriteFloat bstream 0.0 
			
			WriteFloat bstream RestPoseInverse1[2] 
			WriteFloat bstream RestPoseInverse2[2] 
			WriteFloat bstream RestPoseInverse3[2] 
			WriteFloat bstream 0.0
			
			WriteFloat bstream RestPoseInverse1[3] 
			WriteFloat bstream RestPoseInverse2[3] 
			WriteFloat bstream RestPoseInverse3[3] 
			WriteFloat bstream 0.0 
			
			WriteFloat bstream RestPoseInverse4[1] 
			WriteFloat bstream RestPoseInverse4[2] 
			WriteFloat bstream RestPoseInverse4[3] 
			WriteFloat bstream 1.0 

			WriteFloat bstream Position[1] 
			WriteFloat bstream Position[2] 
			WriteFloat bstream Position[3] 
			
			WriteFloat bstream Rotation.w 
			WriteFloat bstream Rotation.x 
			WriteFloat bstream Rotation.y 
			WriteFloat bstream Rotation.z 

			WriteLong bstream nameoffset #unsigned
			WriteLong bstream parentboneindex #unsigned
			
			WriteLong bstream 0 #unsigned
			WriteLong bstream 0 #unsigned
		),
		fn debugOut=
		(
			format "Bone: %\nRestPoseInverse"name
			format "%, %, %, %\n" RestPoseInverse1 RestPoseInverse2 RestPoseInverse3 RestPoseInverse4
			format "Position: %\n" Position
			format "Rotation: %\n" Rotation
			format "nameoffset: %\n" nameoffset
			format "parent: %\n" parentboneindex
		)
	)

	struct e_SCM_Header
	(
		boneoffset,
		wbonecount,    
		vertoffset,
		evertoffset, --omitted in supcom1.0, leave as 0
		vertcount,
		indexoffset, 
		indexcount,
		infooffset, 
		infocount, 
		totalbones,
		
		fn WriteHeader bstream=
		(
			--fourcc "MODL"
			WriteByte bstream 77
			WriteByte bstream 79
			WriteByte bstream 68
			WriteByte bstream 76
			--version number
			WriteLong bstream 5 #unsigned\
			--bonedata offset
			WriteLong bstream boneoffset #unsigned\
			--weighted bone count
			WriteLong bstream wbonecount #unsigned\
			
			--vertex offset
			WriteLong bstream vertoffset #unsigned\
			--extrat vertex offset, always 0
			WriteLong bstream 0 #unsigned\
			--vertex count
			WriteLong bstream vertcount #unsigned\			
			--indexcount
			WriteLong bstream indexoffset #unsigned\
			
			--indexcount
			WriteLong bstream indexcount #unsigned\
			--info offset			
			WriteLong bstream infooffset #unsigned\
			--info count
			WriteLong bstream infocount #unsigned\
			--totalbones			
			WriteLong bstream totalbones #unsigned\
			
			--pad
			for i = 1 to 12 do WriteByte bstream 197
			WriteByte bstream 78
			WriteByte bstream 65
			WriteByte bstream 77
			WriteByte bstream 69
		),
		fn DebugOut=
		(
			format "SCM_Header\n----------\n"
			--format "Marker: %, version: %\n" fourcc version
			format "Weighted bones: %, Total bones: %\n" wbonecount totalbones
			format "Vertex count: %, Trianlge count: %\n" vertcount ( indexcount / 3 )
			format "bOffset: %, vOffset: %, tOffset: %, iOffset %\n" boneoffset vertoffset indexoffset infooffset
			ok
		)
	)
	struct SCM_Export
	(
		bstream,	--bitsream
		bonelist=#(), --array of bone structs
		bonenames=#(), --array of bone names
		wbones=#(),
		vertlist=#(), --array of vert structs
		trislist=#(), --array of tris structs
	
		fn Open filename=
		(
			bstream = fopen filename "wb"
			ok
		),
		
		fn Close=
		(
			fclose bstream
			ok
		),
		fn exportModel=
		(
			--Progress bar
			progressStart ("Exporting Model")
			/*
			if (mod curFrame 3) == 0 then
			(
				progressUpdate (100.0 * max / current)
			)
			*/

			--check if model is selected, has skinmod, etc
			-------------------------------------------------
			progressUpdate (100.0 * 8 / 1)
			
			selection = getCurrentSelection()
			if selection.count == 0 do 
			(
				messageBox "No object selected" title:"error"
				return false
			)
			progressUpdate (100.0 * 8 / 2)
			
			if selection.count > 1 then
			(
				messageBox "Multiple objects selected, please select only 1 object" title:"error"
				return false
			)
			
			progressUpdate (100.0 * 8 / 3)
			select = selection[1]
			object = snapshotasmesh selection[1]

			progressUpdate (100.0 * 8 / 4)
			header = e_SCM_header()
			if object == undefined then
			(
			)

			progressUpdate (100.0 * 8 / 5)
			-------------------------------------------------
			max modify mode
			local skinmod = select.modifiers[#Skin]
			if skinmod == undefined do
			(
				messageBox "No skinmod found" title:"error"
				return false
			)
			progressUpdate (100.0 * 8 / 6)
			-------------------------------------------------
			local num_verts = object.numverts 
			local num_faces = object.numfaces
			header.vertcount = num_verts
			header.indexcount = ( num_faces * 3 )

			progressUpdate (100.0 * 8 / 7)
			totalbones=skinOps.GetNumberBones skinmod
			header.totalbones=totalbones
			boneoffset = 64
			progressUpdate (100.0)
			ProgressEnd()
			
			--make the bone name array
			--progressStart ("Creating Bone List")
			local exportbone = e_SCM_Bone()
			for i = 1 to totalbones do 
			(
				--progressUpdate (100.0 * totalbones / i )
				
				max modify mode
				bonename = skinOps.GetBoneName skinmod i 0
				node = getNodeByName bonename
				exportbone.name = bonename
				transform = node.transform
				--format "Transform: %\n" transform
				exportbone.nameoffset = boneoffset
				parent = node.parent 
				if parent == undefined then
				(
					parentindex = -1
					exportbone.parentboneindex = parentindex
					
					--pTransform = parent.transform
				
					--rTransform = pTransform - transform
				
					exportbone.position = Transform.row4
					exportbone.rotation = quat 0 0 0 1
				)
				else
				(
					parentindex = findItem bonenames parent.name
					parentindex -= 1
				
					exportbone.parentboneindex = parentindex
					pTransform = parent.transform
				
					rTransform = pTransform - transform
				
					exportbone.position = rTransform.row4
					exportbone.rotation = ( rTransform as quat )
				)
				iTransform = inverse transform
				exportbone.RestPoseInverse1 = [ iTransform.row1[1], iTransform.row2[1], iTransform.row3[1]]
				exportbone.RestPoseInverse2 = [ iTransform.row1[2], iTransform.row2[2], iTransform.row3[2]]
				exportbone.RestPoseInverse3 = [ iTransform.row1[3], iTransform.row2[3], iTransform.row3[3]]
				exportbone.RestPoseInverse4 = iTransform.row4
				namelength = ( bonename.count + 1 )
				boneoffset = boneoffset + namelength
				append bonenames bonename
				wbones[i] = 0
				exportbone.debugout()
				append bonelist (copy exportbone)
				--exportbone.writebone bstream
			)
			--ProgressEnd()
			progressStart ("Calculating Header Data")
			----------------------------------------
			--calculate bonename padding
			--var boneoffset should be offset
			modulo = mod boneoffset 16
			bonepadding = 16 - modulo
			if bonepadding < 4 then
			(
				bonepadding += 16
			)
			boneoffset += bonepadding
			bonepadding -= 4
			header.boneoffset = boneoffset
			
			progressUpdate (100.0 * 5 / 1)
			--calculate bonedata padding
			--a single bone entry is 108 bytes. 27*4
			vertexoffset = boneoffset + ( totalbones * 108 )
			modulo = mod vertexoffset 16
			format "mod: %" modulo
			bonedatapadding = 16 - modulo
			format "bdp0: %\n" bonedatapadding
			if bonedatapadding < 4 then
			(
				bonedatapadding += 16
			)
			vertexoffset = ( vertexoffset + bonedatapadding )
			bonedatapadding -= 4
			header.vertoffset = vertexoffset

			progressUpdate (100.0 * 5 / 2)			
			--calculate vertexdata padding
			--a single vertex entry is 68 bytes. 17*4
			indexoffset = vertexoffset + ( num_verts * 68 )
			format "iO: %" indexoffset
			modulo = mod indexoffset 16
			vertdatapadding = 16 - modulo
			if vertdatapadding < 4 then
			(
				vertdatapadding += 16
			)
			indexoffset = ( indexoffset + vertdatapadding )
			format "iO2: %" indexoffset
			vertdatapadding -= 4
			header.indexoffset = indexoffset
			
			progressUpdate (100.0 * 5 / 3)
			--calculate index padding
			--a single index entry is 6 bytes. 3*2
			infooffset = indexoffset + ( num_faces * 6 )
			modulo = mod infooffset 16
			indexpadding = 16 - modulo
			if indexpadding < 4 then
			(
				indexpadding += 16
			)
			infooffset = ( infooffset + indexpadding )
			indexpadding -= 4
			header.infooffset = infooffset
			
			progressUpdate (100.0 * 5 / 4)
			ProgressEnd()
			--progressStart ("Creating Vertex Data")
			--info count
			--fixed string: "Exported with scmtool www.janvanderweg.com"
			header.infocount = 43		
			----------------------------------------
			--get weighted bone count and vertex info
			--for v = 1 to num_verts do
			local exportvert = e_SCM_Vert()
			for v = 1 to num_verts do
			(
				--if (mod v 5) == 0 then
				--(
				--	progressUpdate (100.0 * num_verts / v )
				--)
				max modify mode
				--get position
				exportvert.Position = meshop.getvert object v
				--get vertex normals					
				local vertFaces = meshOp.getFacesUsingVert select v
				local tNormal = [0,0,0]
				for i in vertFaces do tNormal += getFaceNormal select i
				local tNormal = normalize (tNormal / vertFaces.count )
				local bNormal = [0,0,0]
				local tangent = [0,0,0]
				exportvert.Normal = tNormal
				exportvert.Binormal = bNormal
				exportvert.Tangent = tangent
				--uv positions
				exportvert.uv1 = getTVert object v
				exportvert.uv2 = getTVert object v
				/*
				format "Vert: %\n faces: % normal: %\n"v vertfaces tNormal
				local tNormal = [0,0,0]; for i in vertFaces do tNormal += polyOp.getFaceNormal sel i
				local t01 = tNormal / vertFaces.numberSet
				local t02 = normalize (tNormal / vertFaces.numberSet)
				format "My vector: %, Your vector: %\n" t01 t02
				*/
				--Boneindex
				--set boneindex to 0, if a vert isn't weighted it will be linked to the root bone.
				exportvert.boneindex = 0
				weightcount = skinOps.GetVertexWeightCount skinmod v
				for i = 1 to weightcount do
				(
					boneid = skinOps.getvertexweightboneid skinmod v i
					vertweight = skinOps.getvertexweight skinmod v i
					
					if vertweight == 1.0 then
					(
						--format "WEIGHTED - vert: % boneid: % vertweight: %\n" v boneid vertweight
						--This bone should be used
						exportvert.boneindex = (boneid - 1)
						--number of weighted bones
						--Referencepoints don't influence any verts and come after weighted bones in the list.
						--
						header.wbonecount = boneid	
					)
					else
					(
						--discard bone
						--format "vert: % boneid: % vertweight: %\n" v boneid vertweight
					)

						
				)
				--exportvert.debugout()
				append vertlist (copy exportvert)
					--vert = getVert object v
					--format "%," vert 
			)
			local exportface = e_SCM_Tris()
			for f = 1 to num_faces do
			(
				exportface.triindex = getFace object f
				append trislist ( copy exportface )
			)


			header.debugout()
			--ProgressEnd()
			progressStart ("Writing File")

			/*
			-------------------------------------------
			Write scm file
			-------------------------------------------
			*/
			--write header
			header.writeheader bstream
			
			progressUpdate (100.0 * 10 / 1 )
			--write bonelist
			for i = 1 to totalbones do
			(
				Writestring bstream bonenames[i]
			)
			
			progressUpdate (100.0 * 10 / 2 )
			--write padding
			for i = 1 to bonepadding do WriteByte bstream 197
			WriteByte bstream 83
			WriteByte bstream 75
			WriteByte bstream 69
			WriteByte bstream 76
			
			progressUpdate (100.0 * 10 / 3 )
			--write bones
			for i = 1 to totalbones do 
			(
				bonelist[i].writebone bstream
			)
			
			progressUpdate (100.0 * 10 / 4 )
			--write padding
			for i = 1 to bonedatapadding do WriteByte bstream 197
			WriteByte bstream 86
			WriteByte bstream 84
			WriteByte bstream 88
			WriteByte bstream 76
			
			progressUpdate (100.0 * 10 / 5 )
			--write verts
			for i = 1 to num_verts do
			(
				vertlist[i].writevert bstream
				--vertlist[i].debugout
			)

			progressUpdate (100.0 * 10 / 6 )
			--write padding
			for i = 1 to vertdatapadding do WriteByte bstream 197
			WriteByte bstream 84
			WriteByte bstream 82
			WriteByte bstream 73
			WriteByte bstream 83
			
			progressUpdate (100.0 * 10 / 7 )
			--write trislist
			for i = 1 to num_faces do
			(
				trislist[i].writetris bstream
			)
			
			progressUpdate (100.0 * 10 / 8 )
			--write padding
			for i = 1 to indexpadding do WriteByte bstream 197
			WriteByte bstream 73
			WriteByte bstream 78
			WriteByte bstream 70
			WriteByte bstream 79
			
			progressUpdate (100.0 * 10 / 9 )
			--write info
			WriteString bstream "Exported with scmTool www.janvanderweg.com"
			/*
			for v = 1 to num_verts do
			(
				vert = getVert object v
				format "%," vert 
			)
			for f = 1 to num_faces do
			(
				face = getFace object f
				format "%," face 
			)
			*/
			--done creatin header, write it
			--header.debugout()
			--header.writeheader bstream
			--garbage collector
			ProgressEnd()
			gc()
		)
	)
	struct SCM_Loader
	(
		bstream,	--bitsream
		header,		--file header
		bonelist=#(), --array of bone structs
		bonenames=#(), --array of bone names
		vertlist=#(), --array of vert structs
		trislist=#(), --array of tris structs
		vertweight=#(), --array that holds which bone is linked to which vert
		boneweight=#(), --array that holds the bone number
		
		fn Open filename=
		(
			--g_modelFilename = filename
			--g_modelName = getFilenameFile filename
			--g_modelPath = getFileNamePath filename
			bstream = fopen filename "rb"
			ok
		),
		
		fn Close=
		(
			fclose bstream
			--print "Closed filestream"
			ok
		),
		fn LoadModel=
		(
			progressUpdate (0.0)
			
			-- Load SCM_Header
			fseek bstream 0 #seek_set
			header = SCM_Header()
			header.LoadFromStream bstream
			
			progressUpdate (100.0*2/16)
			
			if not header.IsValid() then
			(
				format "Invalid header, aborting.\n"
				return false;
			)
			progressUpdate (100.0*3/16)
			
			-- Load BONES
			fseek bstream 64 #seek_set
			bonestring = ReadFixedString bstream ( header.boneoffset - 64 ) false
			tokens = filterString bonestring "\t"

			--copy all tokens except the last one into bonenames array. Last token is fourcc SKEL.
			for i = 1 to ( tokens.count - 1 ) do
			(
				append bonenames tokens[i]
			)

			fseek bstream header.boneoffset #seek_set
			local curBone = SCM_Bone()
			for i = 1 to bonenames.count do
			(
				curBone.LoadFromStream bstream
				--format "Bone: % - %\n" i bonenames[i]
				--curBone.DebugOut()
				append bonelist (copy curBone)
			)
			progressUpdate (100.0*4/16)

			fseek bstream header.vertoffset #seek_set
			local curVert = SCM_vert()
			for i = 1 to header.vertcount do
			(
				curVert.LoadFromStream bstream
				if i == 1 do curVert.DebugOut()
				append vertlist (copy curVert)
				bone = ( curVert.boneindex + 1 )
				bone = bone[1]
				append boneweight bone
			)
			progressUpdate (100.0*5/16)
			
			fseek bstream header.indexoffset #seek_set
			local curTris = SCM_Tris()
			for i = 1 to header.tricount do
			(
				curTris.LoadFromStream bstream
				append trislist (copy curTris)
			)
			progressUpdate (100.0*6/16)
			
			vert_array=#() -- array that will hold the vertex information
			face_array=#() -- array that will hold the triangle list information
			tvert_array=#() -- array that will hold vertex texture coordinates

			-- build vert array
			for i = 1 to header.vertcount do
			(
				local currentvert = vertlist[i].position
				append vert_array currentvert 
				--use first set of uv's
			    local currenttvert = vertlist[i].uv1
				append tvert_array currenttvert
			)
			progressUpdate (100.0*8/16)
			-- build triangle list
			for i = 1 to header.tricount do
			(
				local currentface = trislist[i].triindex
				currentface[1] = currentface[1] + 1
				currentface[2] = currentface[2] + 1
				currentface[3] = currentface[3] + 1
				append face_array currentface
			)
			progressUpdate (100.0*9/16)
			--create a new mesh object
			new_mesh = mesh vertices:vert_array faces:face_array tverts:tvert_array
			--fix tverts
			fix_mesh = new_mesh.mesh
			BuildTVFaces fix_mesh
			for i = 1 to fix_mesh.numfaces do (setTVFace fix_mesh i (getFace fix_mesh i))
			
			--rebuild the mesh object
			update fix_mesh
			progressUpdate (100.0*10/16)
			--add a skinmodifier to the mesh
			addmodifier new_mesh (skin())
			select new_mesh
			max modify mode
			skinmod=new_mesh.modifiers[1]
			
			progressUpdate (100.0*11/16)
						
			if loadrefpoints.tristate == 1 then
			(
				upto = header.totalbones
			)
			else
			(
				upto = header.wbone
			)
			format "Creating % bones\n" upto
			for i = 1 to upto do
			(
				--inverse bone transformation matrix
				transmatrix = ( inverse bonelist[i].RestPoseInverse )
				xvec = [ transmatrix.row1[1], transmatrix.row2[1], transmatrix.row3[1] ]
				yvec = [ transmatrix.row1[2], transmatrix.row2[2], transmatrix.row3[2] ]
				zvec = [ transmatrix.row1[3], transmatrix.row2[3], transmatrix.row3[3] ]
				format "Bone: % - %\n" i bonenames[i]
				format "xvec: %\nyvec: %\nzvec: %\n" xvec yvec zvec
				local bname = bonenames[i]
				if i > header.wbone then
				(
					if refpointsas.state == 2 then
					(
						--format "---\n%\n" pos
						newdummy = dummy name:bname
						newdummy.boxsize = [2,2,2]
						newdummy.pos = transmatrix.row4
					)
					else
					(
						length = 10 --should find childbone and make length distance between parent and child.
						bonenode = bonesys.createbone transmatrix.row4 (transmatrix.row4 + length*(normalize transmatrix.row1)) (normalize transmatrix.row3)
						--normalize
						--set bone name width etc
						bonenode.name = bname
						bonenode.width = 0.4
						bonenode.height = 0.4
					)
					node = getNodeByName bonenames[i]
					--add bone to skinmod
					if i < upto do skinops.addbone skinmod node 0
					if i == upto do skinops.addbone skinmod node -1
				)
				else
				(
					length = 10 --should find childbone and make length distance between parent and child.
					bonenode = bonesys.createbone transmatrix.row4 (transmatrix.row4 + length*(normalize transmatrix.row1)) (normalize transmatrix.row3)
					--Set bone name width etc
					bonenode.name = bname
					bonenode.width = 0.4
					bonenode.height = 0.4
					
					--Add to skinmod

					node = getNodeByName bonenames[i]
					--add bone to skinmod
					if i < upto do skinops.addbone skinmod node 0
					if i == upto do skinops.addbone skinmod node -1
				)
				
			)
			progressUpdate (100.0*12/16)
			--Fix progressupdate later.
			progressUpdate (100.0*13/16)
			--Weight verts
 			for i = 1 to new_mesh.verts.count do
			(
 				skinops.setvertexweights skinmod i boneweight[i] 1.0
 			)
			
			progressUpdate (100.0*14/16)
			format "Creating % links\n" upto
			for i = 1 to upto do
			(
				--get bone parent
				--print ""
				local parent = bonelist[i].parentboneindex
				local parentname = bonenames[i]
				--format "Parent: %, Name %" parent parentname
				if parent == 0 then
				(
					--do nothing, it's the parent
				)
				else
				(
					--link it to its parent
					node1 = getNodeByName bonenames[i]
					node2 = getNodeByName bonenames[parent]
					attachObjects node2 node1 move:false
					--attachObjects getNodeByName bonenames[i] getNodeByName bonenames[parent] move:false
				)
			)
			
 			progressUpdate (100.0*15/16)
			--fix model rotation
			fixrotnode = getNodeByName bonenames[1]
			if fixrotcheck.tristate == 1 do	
			(
				rotate fixrotnode 90 x_axis
				--switch y and z pos axis of root bone
				fixrotnodepos = fixrotnode.pos
				fixrotnode.pos = [fixrotnodepos[1],fixrotnodepos[3],fixrotnodepos[2]]
			)
			progressUpdate (100.0)
			progressEnd()
		)
	)
	--UI Stuff
	on importButton pressed do
	(
		inipath = scriptsPath + "scmtool.ini"
		filename = getINISetting inipath "settings" "file"
	    scmFilename = getOpenFileName caption:"Select SCM file" \
					filename:filename \
					types:"SupCom SCM (*.scm)|*.scm|All Files (*.*)|*.*|"

		if scmFilename != undefined then
		(
			if DoesFileReallyExist scmFilename then
			(
				progressStart ("Loading Model")
				setINISetting inipath "settings" "file" scmFilename
				
				set coordsys world
				--m3Coordsys = matrix3 [1,0,0] [0,0,1] [0,1,0] [0,0,0]
				--format "m3: %" m3Coordsys
				
				--set coordsys m3Coordsys
				local loader = SCM_Loader()
				loader.Open scmFilename
				progressUpdate (100.0*1/16)
				loader.LoadModel()
				loader.Close()

				max views redraw
			)
		)
	)
	on exportSCM pressed do
	(
		inipath = scriptsPath + "scmtool.ini"
		filename = getINISetting inipath "settings" "exportscm"
	    scmFilename = getSaveFileName caption:"Select SCM file" \
					filename:filename \
					types:"SupCom SCM (*.scm)|*.scm|All Files (*.*)|*.*|"

		--if scmFilename != undefined then
		--(
			--if DoesFileReallyExist scmFilename then
			--(
				setINISetting inipath "settings" "exportscm" scmFilename
				set coordsys world
				--m3Coordsys = matrix3 [1,0,0] [0,0,1] [0,1,0] [0,0,0]
				--set coordsys m3Coordsys

				local exporter = SCM_Export()
				exporter.Open scmFilename
				exporter.exportModel()
				exporter.close()
				
				max views redraw
			--)
		--)

	)

	
)



